"""
Test search classes
"""

import sys
import pytest
import numpy as np

try:
    import xgboost
    from xgboost import XGBClassifier
except ImportError:
    pass

from scipy.stats.distributions import expon
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import RandomForestClassifier

try:
    from skdist.distribute import search
    from skdist.distribute.search import (
        DistGridSearchCV,
        DistRandomizedSearchCV,
        DistMultiModelSearch,
    )

    _import_error = None
except Exception as e:
    _import_error = e


def test_search():
    assert _import_error == None


def test_gs():
    X = np.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]] * 100)
    y = np.array([0, 0, 1] * 100)
    gs = DistGridSearchCV(
        LogisticRegression(solver="liblinear"), {"C": [0.1, 1.0]}, cv=3
    )
    gs.fit(X, y)
    preds = gs.predict(X[:3])
    assert np.allclose(preds, np.array([0, 0, 1]))


def test_rs():
    X = np.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]] * 100)
    y = np.array([0, 0, 1] * 100)
    rs = DistRandomizedSearchCV(
        LogisticRegression(solver="liblinear"), {"C": [0.1, 1.0]}, cv=3, n_iter=2
    )
    rs.fit(X, y)
    preds = rs.predict(X[:3])
    assert np.allclose(preds, np.array([0, 0, 1]))


def test_multimodel():
    X = np.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]] * 100)
    y = np.array([0, 0, 1] * 100)
    models = [
        (
            "rf",
            RandomForestClassifier(n_estimators=10),
            {"max_depth": [5, 10], "min_samples_split": range(2, 10)},
        ),
        (
            "lr0",
            LogisticRegression(multi_class="auto", solver="liblinear"),
            {"C": [0.1, 1.0, 10]},
        ),
        (
            "lr1",
            LogisticRegression(multi_class="auto", solver="liblinear"),
            {"fit_intercept": [True, False], "C": expon()},
        ),
        ("nb", GaussianNB(), {}),
    ]
    clf = DistMultiModelSearch(models, n=2)
    clf.fit(X, y)
    preds = clf.predict(X[:3])
    assert np.allclose(preds, np.array([0, 0, 1]))


@pytest.mark.skipif("xgboost" not in sys.modules, reason="requires xgboost")
def test_fit_params():
    X = np.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]] * 100)
    y = np.array([0, 0, 1] * 100)

    clf = DistRandomizedSearchCV(XGBClassifier(), {"max_depth": [3, 5]}, cv=3, n_iter=2)
    X_test = np.array([[1, 1, 0], [-2, 0, 5], [1, 1, 1]] * 10)
    y_test = np.array([1, 1, 0] * 10)
    fit_params = {
        "eval_metric": "logloss",
        "eval_set": [(X_test, y_test)],
        "early_stopping_rounds": 10,
    }
    clf.fit(X, y, **fit_params)
    preds = clf.predict(X[:3])
    assert np.allclose(preds, np.array([0, 0, 1]))



"""
Test feature eliminator classes
"""

import numpy as np

from scipy.sparse import csr_matrix
from sklearn.linear_model import LogisticRegression
from sklearn import datasets

try:
    from skdist.distribute import eliminate
    from skdist.distribute.eliminate import DistFeatureEliminator

    _import_error = None
except Exception as e:
    _import_error = e


def test_import():
    assert _import_error == None


def test_fit():
    X, y = datasets.load_iris(return_X_y=True)
    X = np.hstack([np.reshape(range(len(X)), (len(X), 1)), X])
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"),
        scoring="f1_weighted",
        min_features_to_select=3,
        cv=3,
    )
    fe.fit(X, y)
    assert np.allclose(fe.best_features_, [1, 2, 3, 4])


def test_score():
    X, y = datasets.load_iris(return_X_y=True)
    X = np.hstack([X, np.reshape(range(len(X)), (len(X), 1))])
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"),
        scoring="f1_weighted",
        min_features_to_select=2,
        cv=3,
    )
    fe.fit(X, y)
    assert round(fe.score(X, y), 2) == 0.96


def test_predict():
    X, y = datasets.load_iris(return_X_y=True)
    X = np.hstack([X, np.reshape(range(len(X)), (len(X), 1))])
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"),
        scoring="f1_weighted",
        min_features_to_select=1,
        cv=3,
        step=2,
    )
    fe.fit(X, y)
    assert np.allclose(fe.best_features_, [0, 1, 2])


def test_sparse():
    X, y = datasets.load_iris(return_X_y=True)
    X = np.hstack([X, np.reshape(range(len(X)), (len(X), 1))])
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"),
        scoring="f1_weighted",
        min_features_to_select=3,
        cv=3,
    )
    fe.fit(csr_matrix(X), y)
    assert fe.predict(X)[0] == 0



"""
Test multiclass classes
"""

try:
    import numpy as np
    from sklearn.linear_model import LogisticRegression
    from skdist.distribute import multiclass
    from skdist.distribute.multiclass import (
        DistOneVsRestClassifier,
        DistOneVsOneClassifier,
    )

    _import_error = None
except Exception as e:
    _import_error = e


def test_multiclass():
    assert _import_error == None


def test_ovr():
    X = np.array([[0, 0, 1, 1], [1, 1, 0, 0], [-1, -1, -1, -1]] * 100)
    y = np.array([0, 1, 2] * 100)
    ovr = DistOneVsRestClassifier(LogisticRegression(solver="liblinear"))
    ovr.fit(X, y)
    preds = ovr.predict(X[:3])
    assert np.allclose(preds, np.array([0, 1, 2]))


def test_ovo():
    X = np.array([[0, 0, 1, 1], [1, 1, 0, 0], [-1, -1, -1, -1]] * 100)
    y = np.array([0, 1, 2] * 100)
    ovo = DistOneVsOneClassifier(LogisticRegression(solver="liblinear"))
    ovo.fit(X, y)
    preds = ovo.predict(X[:3])
    assert np.allclose(preds, np.array([0, 1, 2]))



"""
Various preprocessing classes implemented as scikit-learn
transformers compatible with scikit-learn pipelines.
"""

import warnings
import pandas as pd
import numpy as np

from scipy import sparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import (
    MultiLabelBinarizer,
    FunctionTransformer,
    LabelEncoder,
    normalize,
)
from sklearn.feature_extraction.text import HashingVectorizer
from sklearn import feature_selection

__all__ = [
    "SelectField",
    "FeatureCast",
    "ImputeNull",
    "DenseTransformer",
    "SparseTransformer",
    "LabelEncoderPipe",
    "SelectorMem",
    "HashingVectorizerChunked",
    "MultihotEncoder",
]

_SELECTOR_LOOKUP = {
    "fpr": feature_selection.SelectFpr,
    "fdr": feature_selection.SelectFdr,
    "kbest": feature_selection.SelectKBest,
    "percentile": feature_selection.SelectPercentile,
    "fwe": feature_selection.SelectFwe,
}


class _CustomFunctionTransformer(TransformerMixin, BaseEstimator):
    """ Base class for function transformers """

    def fit(self, X, y=None):
        self.transformer_.fit(X, y=None)
        return self

    def transform(self, X, y=None):
        return self.transformer_.transform(X)


def select_field(X, cols=None, single_dimension=False):
    """
    Select columns from a pandas DataFrame

    Args:
        X (pandas DataFrame): input data
        cols (array-like): list of columns to select
        single_dimension (bool): reduce data to
            one dimension if only one column
            is requested
    Returns:
        X (numpy array)
    """
    if cols is None:
        return X.values
    if len(cols) > 1:
        return X[cols].values
    if len(cols) == 1:
        if single_dimension:
            return X[cols[0]].values
        else:
            return X[cols].values


class SelectField(_CustomFunctionTransformer):
    """
    Applies `select_field` as FunctionTransformer

    Args:
        cols (array-like): list of columns to select
        single_dimension (bool): reduce data to
            one dimension if only one column
            is requested
    """

    def __init__(self, cols=None, single_dimension=False):
        self.cols = cols
        self.single_dimension = single_dimension
        kw_args = {"cols": cols, "single_dimension": single_dimension}
        self.transformer_ = FunctionTransformer(
            select_field, kw_args=kw_args, validate=False
        )


def to_dense(X):
    """ Densify sparse matrix if issparse """
    if sparse.issparse(X):
        return X.todense()
    else:
        return X


class DenseTransformer(_CustomFunctionTransformer):
    """ Applies `to_dense` as a FunctionTransformer """

    def __init__(self):
        self.transformer_ = FunctionTransformer(to_dense, validate=False)


def to_sparse(X):
    """ Sparsify dense matrix if not issparse """
    if sparse.issparse(X):
        return X
    else:
        return sparse.csr_matrix(X)


class SparseTransformer(_CustomFunctionTransformer):
    """ Applies `to_sparse` as a FunctionTransformer """

    def __init__(self):
        self.transformer_ = FunctionTransformer(to_sparse, validate=False)


def feature_cast(X, cast_type=None):
    """
    Casts feature data to requeted type

    Args:
        X (numpy array): input data
        cast_type (type): type to cast data
    Returns:
        X (numpy array)
    """
    if cast_type is None:
        return X
    else:
        return X.astype(cast_type)


class FeatureCast(_CustomFunctionTransformer):
    """
    Applies `feature_cast` as a FunctionTransformer

    Args:
        cast_type (type): type to cast data
    """

    def __init__(self, cast_type=None):
        self.transformer_ = FunctionTransformer(
            feature_cast, kw_args={"cast_type": cast_type}, validate=False
        )


def impute_null(X, impute_val=None):
    """
    Impute null values. Null values are
    determined using `pd.isnull`.

    Args:
        X (numpy array): input data
        impute_val (object): value to impute with
    Returns:
        X (numpy array)
    """
    if impute_val is None:
        return X
    else:
        X[pd.isnull(X)] = impute_val
        return X


class ImputeNull(_CustomFunctionTransformer):
    """
    Applies `impute_null` as a FunctionTransformer

    Args:
        impute_val (object): value to impute with
    """

    def __init__(self, impute_val=None):
        self.transformer_ = FunctionTransformer(
            impute_null, kw_args={"impute_val": impute_val}, validate=False
        )


class LabelEncoderPipe(TransformerMixin, BaseEstimator):
    """
    LabelEncoder wrapper with TransformerMixin. Allows LabelEncoder
    to work with a Pipeline.
    """

    def fit(self, X, y=None):
        """ Fit the LabelEncoder """
        self.le = LabelEncoder()
        self.le.fit(X)
        return self

    def transform(self, X, y=None):
        """ Transform the label encoder """
        return np.array([[x] for x in self.le.transform(X)])


class SelectorMem(TransformerMixin, BaseEstimator):
    """
    Memory efficient feature selector. Identical functionality
    to classes in sklearn.feature_selection.univariate_selection but
    with more memory efficient attribute storage.

    Args:
        selector (object): any sklearn.feature_selection.univariate_selection._BaseFilter
            inheritor
        score_func (function): selector scoring function
        threshold (int/float): scoring threshold to use with scoring function and selector
    """

    def __init__(
        self, selector="fpr", score_func=feature_selection.f_classif, threshold=0.05
    ):
        self.selector = selector
        self.score_func = score_func
        self.threshold = threshold

    def fit(self, X, y=None):
        """
        Run score function on (X, y) and get the appropriate features

        Args:
            X (array-like, shape = [n_samples, n_features])
                The training input samples
            y (array-like, shape = [n_samples])
                The target values (class labels in classification, real numbers in
                regression)
        """
        selector = _SELECTOR_LOOKUP[self.selector.lower()](
            self.score_func, self.threshold
        )
        selector.fit(X, y)
        mask_indices = selector.get_support(indices=True)
        mask_bool = selector.get_support(indices=False)
        if np.array(mask_bool).nbytes > np.array(mask_indices).nbytes:
            self.mask = mask_indices
        else:
            self.mask = mask_bool
        return self

    def transform(self, X, y=None):
        """
        Reduce X to the selected features

        Args:
            X (array of shape [n_samples, n_features])
                The input samples

        Returns:
            X_r (array of shape [n_samples, n_selected_features])
                The input samples with only the selected features
        """
        return X[:, self.mask]


class HashingVectorizerChunked(HashingVectorizer):
    """
    Equivalent to HashingVectorizer but with chunked prediction

    Args:
        chunksize (int): size of transform chunks
    """

    def __init__(self, chunksize=100000, **kwargs):
        self.chunksize = chunksize
        HashingVectorizer.__init__(self, **kwargs)

    def transform(self, X):
        """Transform a sequence of documents to a document-term matrix

        Args:
            X (iterable over raw text documents, length = n_samples)
                Samples. Each sample must be a text document (either bytes or
                unicode strings, file name or file object depending on the
                constructor argument) which will be tokenized and hashed.

        Returns:
            X (scipy.sparse matrix, shape = (n_samples, self.n_features))
                Document-term matrix.
        """
        if (len(X) < self.chunksize) or (self.chunksize is None):
            return self._transform(X)
        else:
            return sparse.vstack([self._transform(x) for x in self._chunks(X)])

    def _transform(self, X):
        if isinstance(X, str):
            raise ValueError(
                "Iterable over raw text documents expected, string object received."
            )

        analyzer = self.build_analyzer()
        X = self._get_hasher().transform(analyzer(doc) for doc in X)
        if self.binary:
            X.data.fill(1)
        if self.norm is not None:
            X = normalize(X, norm=self.norm, copy=False)
        return X

    def _chunks(self, l):
        for i in range(0, len(l), self.chunksize):
            yield l[i : i + self.chunksize]


class MultihotEncoder(BaseEstimator, TransformerMixin):
    """
    Wraps `MultiLabelBinarizer` in a pipeline safe transformer

    Args:
        sparse_output (bool): convert output to sparse matrix
    """

    def __init__(self, sparse_output=False):
        self.transformer = MultiLabelBinarizer()
        self.sparse_output = sparse_output

    def fit(self, X, y=None):
        """ Fit MultiLabelBinarizer """
        self.transformer.fit(X)
        return self

    def transform(self, X, y=None):
        """ Transform MultiLabelBinarizer """
        # ignore unseen label warning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            X_t = self.transformer.transform(X)
        if self.sparse_output:
            return sparse.csr_matrix(X_t)
        else:
            return X_t



"""
Validation functions for the distribute module
"""

import numbers
import numpy as np

from itertools import compress
from scipy.sparse import issparse
from sklearn.model_selection import ParameterGrid
from sklearn.utils.validation import check_is_fitted


def _check_estimator(estimator, verbose=False):
    """ Print sparkContext awareness if apporpriate """
    if verbose:
        if estimator.sc is None:
            print("No spark context is provided; running locally")
        else:
            print("Spark context found; running with spark")


def _check_is_fitted(estimator, attributes=None):
    from sklearn import __version__

    if __version__ < "0.22":
        return check_is_fitted(estimator, attributes)
    else:
        return check_is_fitted(estimator)


def _validate_params(param_sets):
    """ Validates grid params """
    invalid_params = [
        type(param) for param in param_sets if not isinstance(param, dict)
    ]
    if invalid_params:
        raise ValueError(
            "Params must be dictionaries: got " "{0!r}".format(invalid_params)
        )


def _validate_models(models, clf):
    """ Validates models input argument """
    try:
        iter(models)
    except:
        raise TypeError("Input argument 'models' is not iterable.")
    if not (isinstance(models[0], tuple) or isinstance(models[0], list)):
        models = [models]
    names, estimators, param_sets = zip(*models)
    _validate_names(clf, names)
    _validate_estimators(estimators)
    _validate_params(param_sets)
    return models


def _check_base_estimator(estimator):
    """ Validates base estimator """
    return hasattr(estimator, "fit") and hasattr(estimator, "predict")


def _validate_estimators(estimators):
    """ Validates estimators """
    invalid_estimators = [
        type(estimator)
        for estimator in estimators
        if not _check_base_estimator(estimator)
    ]
    if invalid_estimators:
        raise ValueError(
            "Estimators must be sklearn estimators: got "
            "{0!r}".format(invalid_estimators)
        )


def _validate_names(cls, names):
    """ Validates names """
    if len(set(names)) != len(names):
        raise ValueError("Names provided are not unique: " "{0!r}".format(list(names)))
    invalid_names = [name for name in names if not isinstance(name, str)]
    if invalid_names:
        raise ValueError(
            "Estimator names must be strings: got " "{0!r}".format(invalid_names)
        )
    invalid_names = set(names).intersection(cls.get_params(deep=False))
    if invalid_names:
        raise ValueError(
            "Estimator names conflict with constructor "
            "arguments: {0!r}".format(sorted(invalid_names))
        )
    invalid_names = [name for name in names if "__" in name]
    if invalid_names:
        raise ValueError(
            "Estimator names must not contain __: got " "{0!r}".format(invalid_names)
        )


def _check_n_iter(n_iter, param_distributions):
    """
    Check if n_iter is greater than the total number
    of possible param sets from the given distribution.
    """
    all_lists = np.all([not hasattr(v, "rvs") for v in param_distributions.values()])
    if all_lists:
        param_grid = ParameterGrid(param_distributions)
        grid_size = len(param_grid)
    else:
        grid_size = n_iter
    return min(grid_size, n_iter)


def _is_arraylike(x):
    """ Returns whether the input is array-like """
    return hasattr(x, "__len__") or hasattr(x, "shape") or hasattr(x, "__array__")


def _num_samples(x):
    """ Return number of samples in array-like x """
    message = "Expected sequence or array-like, got %s" % type(x)
    if hasattr(x, "fit") and callable(x.fit):
        # Don't get num_samples from an ensembles length!
        raise TypeError(message)

    if not hasattr(x, "__len__") and not hasattr(x, "shape"):
        if hasattr(x, "__array__"):
            x = np.asarray(x)
        else:
            raise TypeError(message)

    if hasattr(x, "shape") and x.shape is not None:
        if len(x.shape) == 0:
            raise TypeError(
                "Singleton array %r cannot be considered" " a valid collection." % x
            )
        # Check that shape is returning an integer or default to len
        # Dask dataframes may not return numeric shape[0] value
        if isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]
    try:
        return len(x)
    except TypeError:
        raise TypeError(message)


def _safe_indexing(X, indices, axis=0):
    """ Return rows, items or columns of X using indices """
    if indices is None:
        return X

    if axis not in (0, 1):
        raise ValueError(
            "'axis' should be either 0 (to index rows) or 1 (to index "
            " column). Got {} instead.".format(axis)
        )

    indices_dtype = _determine_key_type(indices)

    if axis == 0 and indices_dtype == "str":
        raise ValueError("String indexing is not supported with 'axis=0'")

    if axis == 1 and X.ndim != 2:
        raise ValueError(
            "'X' should be a 2D NumPy array, 2D sparse matrix or pandas "
            "dataframe when indexing the columns (i.e. 'axis=1'). "
            "Got {} instead with {} dimension(s).".format(type(X), X.ndim)
        )

    if axis == 1 and indices_dtype == "str" and not hasattr(X, "loc"):
        raise ValueError(
            "Specifying the columns using strings is only supported for "
            "pandas DataFrames"
        )

    if hasattr(X, "iloc"):
        return _pandas_indexing(X, indices, indices_dtype, axis=axis)
    elif hasattr(X, "shape"):
        return _array_indexing(X, indices, indices_dtype, axis=axis)
    else:
        return _list_indexing(X, indices, indices_dtype)


def _determine_key_type(key):
    """ Determine the data type of key """
    err_msg = (
        "No valid specification of the columns. Only a scalar, list or "
        "slice of all integers or all strings, or boolean mask is "
        "allowed"
    )

    dtype_to_str = {int: "int", str: "str", bool: "bool", np.bool_: "bool"}
    array_dtype_to_str = {
        "i": "int",
        "u": "int",
        "b": "bool",
        "O": "str",
        "U": "str",
        "S": "str",
    }

    if key is None:
        return None
    if isinstance(key, tuple(dtype_to_str.keys())):
        try:
            return dtype_to_str[type(key)]
        except KeyError:
            raise ValueError(err_msg)
    if isinstance(key, slice):
        if key.start is None and key.stop is None:
            return None
        key_start_type = _determine_key_type(key.start)
        key_stop_type = _determine_key_type(key.stop)
        if key_start_type is not None and key_stop_type is not None:
            if key_start_type != key_stop_type:
                raise ValueError(err_msg)
        if key_start_type is not None:
            return key_start_type
        return key_stop_type
    if isinstance(key, list):
        unique_key = set(key)
        key_type = {_determine_key_type(elt) for elt in unique_key}
        if not key_type:
            return None
        if len(key_type) != 1:
            raise ValueError(err_msg)
        return key_type.pop()
    if hasattr(key, "dtype"):
        try:
            return array_dtype_to_str[key.dtype.kind]
        except KeyError:
            raise ValueError(err_msg)
    raise ValueError(err_msg)


def _pandas_indexing(X, key, key_dtype, axis):
    """ Index a pandas dataframe or a series """
    if hasattr(key, "shape"):
        # Work-around for indexing with read-only key in pandas
        # FIXME: solved in pandas 0.25
        key = np.asarray(key)
        key = key if key.flags.writeable else key.copy()
    # check whether we should index with loc or iloc
    indexer = X.iloc if key_dtype == "int" else X.loc
    return indexer[:, key] if axis else indexer[key]


def _list_indexing(X, key, key_dtype):
    """ Index a Python list """
    if np.isscalar(key) or isinstance(key, slice):
        # key is a slice or a scalar
        return X[key]
    if key_dtype == "bool":
        # key is a boolean array-like
        return list(compress(X, key))
    # key is a integer array-like of key
    return [X[idx] for idx in key]


def _array_indexing(array, key, key_dtype, axis):
    """ Index an array or scipy.sparse consistently across NumPy version """
    if issparse(array):
        if key_dtype == "bool":
            key = np.asarray(key)
    return array[key] if axis == 0 else array[:, key]



"""
Utility functions for the distribute module
"""

import numbers
import numpy as np

from sklearn.metrics import check_scoring

from .validation import _safe_indexing
from abc import ABCMeta, abstractmethod
from operator import attrgetter
from functools import update_wrapper
import numpy as np

def _aggregate_score_dicts(scores):
    """ Aggregate the list of dict to dict of np ndarray """
    return {key: np.asarray([score[key] for score in scores]) for key in scores[0]}


def _multimetric_score(estimator, X_test, y_test, scorers):
    """ Return a dict of score for multimetric scoring """
    scores = {}

    for name, scorer in scorers.items():
        if y_test is None:
            score = scorer(estimator, X_test)
        else:
            score = scorer(estimator, X_test, y_test)

        if hasattr(score, "item"):
            try:
                # e.g. unwrap memmapped scalars
                score = score.item()
            except ValueError:
                # non-scalar?
                pass
        scores[name] = score

        if not isinstance(score, numbers.Number):
            raise ValueError(
                "scoring must return a number, got %s (%s) "
                "instead. (scorer=%s)" % (str(score), type(score), name)
            )
    return scores


def _score(estimator, X_test, y_test, scorer, is_multimetric=False):
    """
    Compute the score(s) of an estimator on a given test set. Will return
    a single float if is_multimetric is False and a dict of floats,
    if is_multimetric is True
    """
    if is_multimetric:
        return _multimetric_score(estimator, X_test, y_test, scorer)
    else:
        if y_test is None:
            score = scorer(estimator, X_test)
        else:
            score = scorer(estimator, X_test, y_test)

        if hasattr(score, "item"):
            try:
                # e.g. unwrap memmapped scalars
                score = score.item()
            except ValueError:
                # non-scalar?
                pass

        if not isinstance(score, numbers.Number):
            raise ValueError(
                "scoring must return a number, got %s (%s) "
                "instead. (scorer=%r)" % (str(score), type(score), scorer)
            )
    return score


def _check_multimetric_scoring(estimator, scoring=None):
    """ Check the scoring parameter in cases when multiple metrics are allowed """
    if callable(scoring) or scoring is None or isinstance(scoring, str):
        scorers = {"score": check_scoring(estimator, scoring=scoring)}
        return scorers, False
    else:
        err_msg_generic = (
            "scoring should either be a single string or "
            "callable for single metric evaluation or a "
            "list/tuple of strings or a dict of scorer name "
            "mapped to the callable for multiple metric "
            "evaluation. Got %s of type %s" % (repr(scoring), type(scoring))
        )

        if isinstance(scoring, (list, tuple, set)):
            err_msg = (
                "The list/tuple elements must be unique "
                "strings of predefined scorers. "
            )
            invalid = False
            try:
                keys = set(scoring)
            except TypeError:
                invalid = True
            if invalid:
                raise ValueError(err_msg)

            if len(keys) != len(scoring):
                raise ValueError(
                    err_msg + "Duplicate elements were found in"
                    " the given list. %r" % repr(scoring)
                )
            elif len(keys) > 0:
                if not all(isinstance(k, str) for k in keys):
                    if any(callable(k) for k in keys):
                        raise ValueError(
                            err_msg + "One or more of the elements were "
                            "callables. Use a dict of score name "
                            "mapped to the scorer callable. "
                            "Got %r" % repr(scoring)
                        )
                    else:
                        raise ValueError(
                            err_msg + "Non-string types were found in "
                            "the given list. Got %r" % repr(scoring)
                        )
                scorers = {
                    scorer: check_scoring(estimator, scoring=scorer)
                    for scorer in scoring
                }
            else:
                raise ValueError(err_msg + "Empty list was given. %r" % repr(scoring))

        elif isinstance(scoring, dict):
            keys = set(scoring)
            if not all(isinstance(k, str) for k in keys):
                raise ValueError(
                    "Non-string types were found in the keys of "
                    "the given dict. scoring=%r" % repr(scoring)
                )
            if len(keys) == 0:
                raise ValueError("An empty dict was passed. %r" % repr(scoring))
            scorers = {
                key: check_scoring(estimator, scoring=scorer)
                for key, scorer in scoring.items()
            }
        else:
            raise ValueError(err_msg_generic)
        return scorers, True


def _num_samples(x):
    """ Return number of samples in array-like x """
    if hasattr(x, "fit") and callable(x.fit):
        # Don't get num_samples from an ensembles length!
        raise TypeError("Expected sequence or array-like, got " "estimator %s" % x)
    if not hasattr(x, "__len__") and not hasattr(x, "shape"):
        if hasattr(x, "__array__"):
            x = np.asarray(x)
        else:
            raise TypeError("Expected sequence or array-like, got %s" % type(x))
    if hasattr(x, "shape"):
        if len(x.shape) == 0:
            raise TypeError(
                "Singleton array %r cannot be considered" " a valid collection." % x
            )
        # Check that shape is returning an integer or default to len
        # Dask dataframes may not return numeric shape[0] value
        if isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]
        else:
            return len(x)
    else:
        return len(x)


def _safe_split(estimator, X, y, indices, train_indices=None):
    """Create subset of dataset and properly handle kernels"""
    from sklearn.gaussian_process.kernels import Kernel as GPKernel

    if (
        hasattr(estimator, "kernel")
        and callable(estimator.kernel)
        and not isinstance(estimator.kernel, GPKernel)
    ):
        # cannot compute the kernel values with custom function
        raise ValueError(
            "Cannot use a custom kernel function. "
            "Precompute the kernel matrix instead."
        )

    if not hasattr(X, "shape"):
        if getattr(estimator, "_pairwise", False):
            raise ValueError(
                "Precomputed kernels or affinity matrices have "
                "to be passed as arrays or sparse matrices."
            )
        X_subset = [X[index] for index in indices]
    else:
        if getattr(estimator, "_pairwise", False):
            # X is a precomputed square kernel matrix
            if X.shape[0] != X.shape[1]:
                raise ValueError("X should be a square kernel matrix")
            if train_indices is None:
                X_subset = X[np.ix_(indices, indices)]
            else:
                X_subset = X[np.ix_(indices, train_indices)]
        else:
            X_subset = _safe_indexing(X, indices)

    if y is not None:
        y_subset = _safe_indexing(y, indices)
    else:
        y_subset = None
    return X_subset, y_subset


def _dict_slice_remove(a, b, cols=["model_index", "param_set"]):
    """
    Remove dictionary records from list 'a'
    if the records appear in 'b'
    for the given columns 'cols'.
    """
    b_ = [{k: d[k] for k in cols} for d in b]
    a_ = []
    for d in a:
        if {k: d[k] for k in cols} not in b_:
            a_.append(d)
    return a_

class _IffHasAttrDescriptor:
    """Implements a conditional property using the descriptor protocol.

    Using this class to create a decorator will raise an ``AttributeError``
    if none of the delegates (specified in ``delegate_names``) is an attribute
    of the base object or the first found delegate does not have an attribute
    ``attribute_name``.

    This allows ducktyping of the decorated method based on
    ``delegate.attribute_name``. Here ``delegate`` is the first item in
    ``delegate_names`` for which ``hasattr(object, delegate) is True``.

    See https://docs.python.org/3/howto/descriptor.html for an explanation of
    descriptors.
    """
    def __init__(self, fn, delegate_names, attribute_name):
        self.fn = fn
        self.delegate_names = delegate_names
        self.attribute_name = attribute_name

        # update the docstring of the descriptor
        update_wrapper(self, fn)

    def __get__(self, obj, type=None):
        # raise an AttributeError if the attribute is not present on the object
        if obj is not None:
            # delegate only on instances, not the classes.
            # this is to allow access to the docstrings.
            for delegate_name in self.delegate_names:
                try:
                    delegate = attrgetter(delegate_name)(obj)
                except AttributeError:
                    continue
                else:
                    getattr(delegate, self.attribute_name)
                    break
            else:
                attrgetter(self.delegate_names[-1])(obj)

        # lambda, but not partial, allows help() to work with update_wrapper
        out = lambda *args, **kwargs: self.fn(obj, *args, **kwargs)
        # update the docstring of the returned function
        update_wrapper(out, self.fn)
        return out

def if_delegate_has_method(delegate):
    """Create a decorator for methods that are delegated to a sub-estimator

    This enables ducktyping by hasattr returning True according to the
    sub-estimator.

    Parameters
    ----------
    delegate : string, list of strings or tuple of strings
        Name of the sub-estimator that can be accessed as an attribute of the
        base object. If a list or a tuple of names are provided, the first
        sub-estimator that is an attribute of the base object will be used.

    """
    if isinstance(delegate, list):
        delegate = tuple(delegate)
    if not isinstance(delegate, tuple):
        delegate = (delegate,)

    return lambda fn: _IffHasAttrDescriptor(fn, delegate,
                                            attribute_name=fn.__name__)



"""
Test postprocessing 
"""

try:
    import numpy as np
    from sklearn.linear_model import LogisticRegression
    from skdist import postprocessing

    _import_error = None
except Exception as e:
    _import_error = e


def test_import_postprocessing():
    assert _import_error == None


def test_predict():
    X = np.array([[1, 2, 3], [4, 5, 6]])
    y = np.array([0, 1])

    model1 = LogisticRegression(solver="liblinear")
    model2 = LogisticRegression(solver="lbfgs")
    model1.fit(X, y)
    model2.fit(X, y)

    clf = postprocessing.SimpleVoter(
        [("model1", model1), ("model2", model2)], voting="soft", classes=model1.classes_
    )
    pred = clf.predict(X)
    probs = clf.predict_proba(X)
    assert pred.shape == y.shape


def test_predict_hard_voting():
    X = np.array([[1, 2, 3], [4, 5, 6]])
    y = np.array([0, 1])

    model1 = LogisticRegression(solver="liblinear")
    model2 = LogisticRegression(solver="lbfgs")
    model1.fit(X, y)
    model2.fit(X, y)

    clf = postprocessing.SimpleVoter(
        [("model1", model1), ("model2", model2)], voting="hard", classes=model1.classes_
    )
    pred = clf.predict(X)
    assert np.allclose(pred, y)


def test_predict_strings():
    X = np.array([[1, 2, 3], [4, 5, 6]])
    y = np.array(["pizza", "tacos"])

    model1 = LogisticRegression(solver="liblinear")
    model2 = LogisticRegression(solver="lbfgs")
    model1.fit(X, y)
    model2.fit(X, y)

    clf = postprocessing.SimpleVoter(
        [("model1", model1), ("model2", model2)], voting="hard", classes=model1.classes_
    )
    pred = clf.predict(X)
    assert list(pred) == list(y)



"""
Default feature encoding functions and pipelines
for automated feature transformation.
"""

from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction import DictVectorizer
from sklearn.feature_selection import VarianceThreshold
from sklearn.impute import SimpleImputer

from ..preprocessing import (
    ImputeNull,
    SelectField,
    FeatureCast,
    LabelEncoderPipe,
    HashingVectorizerChunked,
    MultihotEncoder,
)


def tokenizer(x):
    """ Trivial tokenizer """
    return x


dict_encoder = lambda c: [
    (
        "{0}_dict_encoder".format(c),
        Pipeline(
            steps=[
                ("var", SelectField(cols=[c], single_dimension=True)),
                ("fillna", ImputeNull({})),
                ("vec", DictVectorizer()),
            ]
        ),
    )
]

onehot_encoder = lambda c: [
    (
        "{0}_onehot".format(c),
        Pipeline(
            steps=[
                ("var", SelectField(cols=[c], single_dimension=True)),
                ("cast", FeatureCast(cast_type=str)),
                ("fillna", ImputeNull("")),
                (
                    "vec",
                    CountVectorizer(
                        token_pattern=None,
                        tokenizer=tokenizer,
                        binary=True,
                        decode_error="ignore",
                    ),
                ),
            ]
        ),
    )
]

multihot_encoder = lambda c: [
    (
        "{0}_multihot".format(c),
        Pipeline(
            steps=[
                ("var", SelectField(cols=[c], single_dimension=True)),
                ("fillna", ImputeNull([])),
                ("vec", MultihotEncoder()),
            ]
        ),
    )
]

numeric_encoder = lambda c: [
    (
        "{0}_scaler".format(c),
        Pipeline(
            steps=[
                ("var", SelectField(cols=[c])),
                ("imputer", SimpleImputer(strategy="median")),
                ("scaler", StandardScaler(copy=False)),
            ]
        ),
    )
]

_default_encoders = {
    "small": {
        "string_vectorizer": lambda c: [
            (
                "{0}_word_vec".format(c),
                Pipeline(
                    steps=[
                        ("var", SelectField(cols=[c], single_dimension=True)),
                        ("fillna", ImputeNull(" ")),
                        (
                            "vec",
                            HashingVectorizerChunked(
                                ngram_range=(1, 2),
                                analyzer="word",
                                decode_error="ignore",
                            ),
                        ),
                        ("var_thresh", VarianceThreshold()),
                    ]
                ),
            )
        ],
        "onehotencoder": onehot_encoder,
        "multihotencoder": multihot_encoder,
        "numeric": numeric_encoder,
        "dict": dict_encoder,
    },
    "medium": {
        "string_vectorizer": lambda c: [
            (
                "{0}_word_vec".format(c),
                Pipeline(
                    steps=[
                        ("var", SelectField(cols=[c], single_dimension=True)),
                        ("fillna", ImputeNull(" ")),
                        (
                            "vec",
                            HashingVectorizerChunked(
                                ngram_range=(1, 3),
                                analyzer="word",
                                decode_error="ignore",
                            ),
                        ),
                        ("var_thresh", VarianceThreshold()),
                    ]
                ),
            ),
            (
                "{0}_char_vec".format(c),
                Pipeline(
                    steps=[
                        ("var", SelectField(cols=[c], single_dimension=True)),
                        ("fillna", ImputeNull(" ")),
                        (
                            "vec",
                            HashingVectorizerChunked(
                                ngram_range=(3, 4),
                                analyzer="char_wb",
                                decode_error="ignore",
                            ),
                        ),
                        ("var_thresh", VarianceThreshold()),
                    ]
                ),
            ),
        ],
        "onehotencoder": onehot_encoder,
        "multihotencoder": multihot_encoder,
        "numeric": numeric_encoder,
        "dict": dict_encoder,
    },
    "large": {
        "string_vectorizer": lambda c: [
            (
                "{0}_word_vec".format(c),
                Pipeline(
                    steps=[
                        ("var", SelectField(cols=[c], single_dimension=True)),
                        ("fillna", ImputeNull(" ")),
                        (
                            "vec",
                            HashingVectorizerChunked(
                                ngram_range=(1, 3),
                                analyzer="word",
                                decode_error="ignore",
                            ),
                        ),
                        ("var_thresh", VarianceThreshold()),
                    ]
                ),
            ),
            (
                "{0}_char_vec".format(c),
                Pipeline(
                    steps=[
                        ("var", SelectField(cols=[c], single_dimension=True)),
                        ("fillna", ImputeNull(" ")),
                        (
                            "vec",
                            HashingVectorizerChunked(
                                ngram_range=(2, 5),
                                analyzer="char_wb",
                                decode_error="ignore",
                            ),
                        ),
                        ("var_thresh", VarianceThreshold()),
                    ]
                ),
            ),
        ],
        "onehotencoder": onehot_encoder,
        "multihotencoder": multihot_encoder,
        "numeric": numeric_encoder,
        "dict": dict_encoder,
    },
}



"""
Test encoder classes
"""

try:
    import pandas as pd
    import numpy as np
    from skdist.distribute import encoder
    from skdist.distribute.encoder import Encoderizer, EncoderizerExtractor

    _import_error = None
except Exception as e:
    _import_error = e


def test_encoder():
    assert _import_error == None


def test_simple():
    # simple encoderizer example
    X = [
        "text encoding is fun",
        "this is a text encoding example",
        "hopefully it works and we'll have passed the test",
    ]
    encoderizer = encoder.Encoderizer(col_names=["text"])
    X_t = encoderizer.fit_transform(X)
    assert X_t.shape == (3, 31)


def test_pandas():
    # pandas encoderizer example
    X = [
        "text encoding is fun",
        "this is a text encoding example",
        "hopefully it works and we'll have passed the test",
    ]
    df = pd.DataFrame(
        {
            "text": X * 4,
            "constant_int": [np.nan, 2, 2] * 4,
            "constant_str": ["test", "test", "test"] * 4,
            "constant_null": [None, "test", np.nan] * 4,
            "numbers": [0.123, None, 0.535] * 4,
            "dicts": [{"a": 4}, None, {"b": 1}] * 4,
        }
    )
    encoderizer = encoder.Encoderizer(size="medium")
    X_t = encoderizer.fit_transform(df)
    assert X_t.shape == (12, 180)


def test_dict_list():
    # pandas encoderizer example
    X = [
        "[text encoding is fun]",
        "this is a text encoding {'pizza': 'example'}",
        "hopefully it works and [we'll have passed] the test",
    ]
    df = pd.DataFrame(
        {
            "text": X * 4,
            "constant_str": ["test", "test", "test"] * 4,
            "dicts": [{"a": 4}, None, {"b": 1}] * 4,
            "lists": [["this", "is", "text"], ["more", "text"], ["text"]] * 4,
        }
    )
    encoderizer = encoder.Encoderizer(size="medium")
    X_t = encoderizer.fit_transform(df)
    assert X_t.shape == (12, 210)


def test_dict_list_tuple():
    # pandas encoderizer example
    X = [
        "[text encoding is fun]",
        "this is a text encoding {'pizza': 'example'}",
        "hopefully it works and [we'll have passed] the test",
    ]
    df = pd.DataFrame(
        {
            "text_col": X * 4,
            "categorical_str_col": ["control", "treatment", "control"] * 4,
            "categorical_int_col": [0, 1, 2] * 4,
            "numeric_col": [5, 22, 69] * 4,
            "dict_col": [{"a": 4}, {"b": 1}, {"c": 3}] * 4,
            "list_col": [[1, 2], [1, 3], [2]] * 4,
            "tuple_col": [(1, 2), (1, 3), (2,)] * 4,
        }
    )
    encoderizer = encoder.Encoderizer(size="medium")
    X_t = encoderizer.fit_transform(df)
    assert X_t.shape == (12, 244)



"""
Run setup
"""

from setuptools import setup, find_packages
from skdist import __version__

DISTNAME = "sk-dist-ems"
VERSION = __version__
DESCRIPTION = "Distributed scikit-learn meta-estimators with PySpark"
with open("README.rst") as f:
    LONG_DESCRIPTION = f.read()
CLASSIFIERS = [
    "Intended Audience :: Developers",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
]
AUTHOR = "Ibotta Inc."
AUTHOR_EMAIL = "machine_learning@ibotta.com"
LICENSE = "Apache 2.0"
DOWNLOAD_URL = "https://pypi.org/project/sk-dist/#files"
PROJECT_URLS = {"Source Code": "https://github.com/EricSchles/sk-dist"}
MIN_PYTHON_VERSION = "3.10"
MIN_PANDAS_VERSION = "0.17.0"
MIN_NUMPY_VERSION = "2.2.5"
MIN_SKLEARN_VERSION = "1.6.0"
MAX_SKLEARN_VERSION = "1.8.0"
MIN_XGBOOST_VERSION = "0.4"
PYARROW_VERSION = "21.0.0"
MIN_PYSPARK_VERSION = "4.0.1"
MIN_PYTESTSPARK_VERSION = "0.4.5"

install_requires = [
    "scikit-learn>={0},<{1}".format(MIN_SKLEARN_VERSION, MAX_SKLEARN_VERSION),
    "pandas>={0}".format(MIN_PANDAS_VERSION),
    "numpy>={0}".format(MIN_NUMPY_VERSION),
    "scipy",
    "joblib",
]

tests_require = [
    "xgboost>={0}".format(MIN_XGBOOST_VERSION),
    "pyarrow=={0}".format(PYARROW_VERSION),
    "pyspark>={0}".format(MIN_PYSPARK_VERSION),
    "pytest-spark>={0}".format(MIN_PYTESTSPARK_VERSION),
]


def parse_description(description):
    """
    Strip figures and alt text from description
    """
    return "\n".join(
        [
            a
            for a in description.split("\n")
            if ("figure::" not in a) and (":alt:" not in a)
        ]
    )


setup(
    name=DISTNAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=parse_description(LONG_DESCRIPTION),
    classifiers=CLASSIFIERS,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    download_url=DOWNLOAD_URL,
    project_urls=PROJECT_URLS,
    packages=find_packages(),
    python_requires=">={0}".format(MIN_PYTHON_VERSION),
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=dict(tests=tests_require),
)



"""
Various postprocessing classes implemented as scikit-learn
transformers compatible with scikit-learn pipelines.
"""

import numpy as np

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import Bunch

from .distribute.validation import _check_is_fitted

__all__ = ["SimpleVoter"]


class SimpleVoter(BaseEstimator, ClassifierMixin):
    """
    Simple voting classifier which takes fitted estimators as
    input and applies voting ensembling at predict time. Similar to
    `sklearn.enseble.VotingClassifier` except the VotingClassifier
    will actually fit the child estimators, where the SimpleVoter
    will take fitted estimators at instantiation and apply similar
    prediction functions as predict time.

    The VotingClassifier implements a dummy fit method that does nothing
    unless the class attributes `estimators` or `classes` have been
    manually updated.

    The value add over `sklearn.enseble.VotingClassifier` is that
    SimpleVoter can solely run the predict, and let the fit live elsewhere.

    Args:
        estimators (list of tuples): list of fitted (name, estimator)
            tuples for voting
        classes (array-like): list of class names from one
            of the underlying estimators
        voting (str): voting method ('hard' or 'soft')
        weights (array-like): array of weights corresponding
            to each estimator

    """

    def __init__(self, estimators, classes, voting="hard", weights=None):
        self.estimators = estimators
        self.classes = classes
        self.voting = voting
        self.weights = weights
        self._assemble_attributes()

    @property
    def named_estimators(self):
        """ Bunches the estimators by name """
        return Bunch(**dict(self.estimators))

    @property
    def _weights_not_none(self):
        """Get the weights of not `None` estimators"""
        if self.weights is None:
            return None
        return [
            w
            for est, w in zip(self.estimators, self.weights)
            if est[1] not in (None, "drop")
        ]

    def fit(self, X, y=None):
        """ Trivial fit method; re-assembles attributes """
        self._assemble_attributes()
        return self

    def predict(self, X):
        """ Compute predictions for samples in X """
        _check_is_fitted(self, "estimators_")
        if self.voting == "soft":
            maj = np.argmax(self.predict_proba(X), axis=1)
        else:  # 'hard' voting
            predictions = self._predict(X)
            maj = np.apply_along_axis(
                lambda x: np.argmax(np.bincount(x, weights=self._weights_not_none)),
                axis=1,
                arr=predictions,
            )
        maj = self.le_.inverse_transform(maj)
        return maj

    @property
    def predict_proba(self):
        """ Compute probabilities of possible outcomes for samples in X """
        return self._predict_proba

    def _predict(self, X):
        "" "Collect results from clf.predict calls " ""
        return np.asarray(
            [self.le_.transform(clf.predict(X)) for clf in self.estimators_]
        ).T

    def _predict_proba(self, X):
        """ Predict class probabilities for X in 'soft' voting """
        if self.voting == "hard":
            raise AttributeError(
                "predict_proba is not available when voting=%r" % self.voting
            )
        _check_is_fitted(self, "estimators_")
        avg = np.average(
            self._collect_probas(X), axis=0, weights=self._weights_not_none
        )
        return avg

    def _collect_probas(self, X):
        """Collect results from clf.predict calls. """
        return np.asarray([clf.predict_proba(X) for clf in self.estimators_])

    def _assemble_attributes(self):
        """ Assemble fitted class attributes """
        names, clfs = zip(*self.estimators)
        self.estimators_ = clfs
        self.classes_ = self.classes
        self.le_ = LabelEncoder()
        self.le_.classes_ = self.classes
